//
//  Copyright 2010 Xilinx, Inc.
//
//  Licensed under the Apache License, Version 2.0 (the "License");
//  you may not use this file except in compliance with the License.
//  You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
//  Unless required by applicable law or agreed to in writing, software
//  distributed under the License is distributed on an "AS IS" BASIS,
//  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//  See the License for the specific language governing permissions and
//  limitations under the License.
//
//  -----------------------------------------------------------------------
//
//  File:        runtime_verbosity_pkg.sv
//  Revision:    1.0, 2010-Nov-05
//  Description: Package to provide runtime, component-level control over
//               verbosity in an OVM testbench.
//  Usage:
//    1) Import this package in your TB (preferrably the top-level)
//    2) From each component which is to be controlled, call
//       runtime_verbosity_pkg::set_verbosity(this, <report_id>)
//       in the build() function. If there are any ovm_report_info or
//       `ovm_info calls in build(), call set_verbosity() _before_ the *_info
//       calls. If the component has sub-components, call set_verbosity()
//       _after_ the sub-components are created. If there are *_info calls
//       _and_ sub-components, call set_verbosity() twice - once before the
//       *_info and instantiations, and once at the very end of build(). This
//       is necessary for recursive verbosity control to work.
//    3) Control verbosity by passing in arguments to your simulation
//       command-line. Syntax is as follows:
//         <sim> +VERB_OVM_<level>=[^]<identifier>[:[^]<identifier>][...]
//       where:
//         <sim> is your simulation command
//         <level> is NONE, LOW, MEDIUM, HIGH, FULL, or DEBUG
//         <identifier> is either the id used in calls to ovm_report_info
//           or a '~' followed by the hierarchical path to a component
//         The '^' indicates that this verbosity setting should be applied
//           recursively to the component and all its children
//       Note that there may be 0 or 1 +VERB_OVM_LEVEL plusarg for each
//         of the possible verbosity levels
//

`include "ovm_macros.svh"

package runtime_verbosity_pkg;

   import ovm_pkg::*;

   // A packed struct to allow us to only do a single associative array lookup
   typedef struct packed {
                          int level;
                          bit recur;
                          } verbosity_t;

   // Associative array to hold verbosity setting for each report_id
   verbosity_t verbosity[string];
   // Associative array to hold verbosity setting for each hierarchical name
   verbosity_t hverbosity[string];

   // Utility class which is used to parse the plusargs and build the
   // lookup table for overriding verbosity at runtime.
   // Note that this class is instantiated as a singleton in this package,
   // which means an instance will automatically get created (and its 'new'
   // function called) if the package is used anywhere
   class verbosity_plusargs extends ovm_object;
      string report_id = "VERBOSITY_PLUSARGS";

      function new(string name = "verbosity_plusargs");
         super.new(name);

         parse_plusargs();
      endfunction : new

      // Function to parse the command-line plusargs (at runtime) and populate
      // the lookup-table. Once the table is populated, it is added to the
      // configuration database (at the ovm_top level). Note that the actual
      // work of parsing the command-line is done in the parse_verbosities
      // function.
      virtual function void parse_plusargs();
         // Check plusargs for each verbosity level in turn. Ordered from
         // least-verbose to most-verbose, so higher verbosity will override
         // lower, if set in two places
         parse_verbosities("NONE",   OVM_NONE);
         parse_verbosities("LOW",    OVM_LOW);
         parse_verbosities("MEDIUM", OVM_MEDIUM);
         parse_verbosities("HIGH",   OVM_HIGH);
         parse_verbosities("FULL",   OVM_FULL);
         parse_verbosities("DEBUG",  OVM_DEBUG);
      endfunction : parse_plusargs

      // Function to parse command-line verbosity arguments.
      // Entries can either be a single value, or multiple values
      // separated by colons. For each value, a configuration entry is made
      // with the corresponding verbosity setting. Each value can be a
      // report_id or a hierarchical name, and can optionally set the
      // verbosity for all child objects if prefaced with "^". Hierarchical
      // names are prefaced by "~".
      virtual function void parse_verbosities(string level_string,
                                              int level_int);
         string  verb_string, tmp_string;
         int     left_pos = 0, right_pos = 0;
         int     str_len;
         bit     is_hier = 0;
         bit     is_recur = 0;
         bit     is_first = 1;
         verbosity_t tmp_verbosity;

         // Query plusargs for the verbosity level specified in level_string
         if ($value$plusargs({"VERB_OVM_", level_string, "=%s"},
                             verb_string)) begin
            `ovm_info(report_id,
                      {"Found verbosity setting VERB_OVM_",
                       level_string, "=", verb_string},
                      OVM_DEBUG)
            str_len = verb_string.len();

            if (verb_string[str_len-1] != ":") begin
              verb_string = {verb_string, ":"};
              str_len++;
            end

            // Loop through the string, looking for a ':' or end-of-string
            do begin

               case (verb_string[right_pos])
                  "~": begin // Check for a "~" (hierarchical name) qualifier
                     if (!is_first) begin
                       `ovm_error(report_id,
                          "Found ~ qualifier after beginning of string");
                     end
                     is_hier = 1;
                     left_pos++; // The "~" is not part of the resulting string
                     right_pos++; // Move search position to the right
                  end // "~"
                  "^": begin // Check for a "^" (recursive) qualifier
                     if (!is_first) begin
                       `ovm_error(report_id,
                          "Found ^ qualifier after beginning of string");
                     end
                     is_recur = 1;
                     left_pos++; // The "^" is not part of the resulting string
                     right_pos++; // Move search position to the right
                  end // "^"
                  ":": begin // Check for a ":" (separator)
                     // If the separator was found, a report_id or hierarchical
                     // name was found between the last separator (if any) and
                     // the current position - 1. Extract the string and store
                     // it in the lookup-table, along with the the value
                     // corresponding to the specified level_string. If the
                     // string contained a "." it is considered a hierarchical
                     // name; if not, it is considered a report_id
                     tmp_string = verb_string.substr(left_pos, right_pos-1);
                     `ovm_info(report_id,
                        {is_hier ? "Hierarchical verbosity" : "Verbosity",
                           is_recur ? " (recursive)" : "",
                           " list entry: ",
                           tmp_string,
                           " = OVM_",
                           level_string},
                        OVM_DEBUG)
                     if (is_hier) begin
                       tmp_verbosity.level = level_int;
                       tmp_verbosity.recur = is_recur;
                       hverbosity[tmp_string] = tmp_verbosity;
                       is_hier = 0;
                       is_recur = 0;
                       is_first = 1;
                     end // if (is_hier)
                     else begin
                       tmp_verbosity.level = level_int;
                       tmp_verbosity.recur = is_recur;
                       verbosity[tmp_string] = tmp_verbosity;
                       is_hier = 0;
                       is_recur = 0;
                       is_first = 1;
                     end

                     // Update the left (stored) position to the character
                     // after the separator, and the right (search) position
                     // to the character that
                     left_pos  = right_pos + 1;
                     right_pos = left_pos;
                  end // ":"
                  default: begin // Any other non-special character
                     // If no qualifier or separator was found, keep moving
                     // the search position to the right
                     right_pos++;
                     is_first = 0; // No qualifiers allowed once a
                                   // non-qualifier is detected
                  end // default
               endcase // verb_string[right_pos]
            end while (right_pos < str_len);
         end // if ($value$plusargs({"VERB_OVM_", level_string, "=%s"},
             //     verb_string))
      endfunction : parse_verbosities

   endclass : verbosity_plusargs

   // Create singleton utility class which will parse the plusargs at runtime
   verbosity_plusargs verb_plusargs = new();

   // Function which can be called by components to set their verbosity
   // threshold from the lookup-table. First, check for report_id, then check
   // for hierarchical name - this gives hierarchical name-based verbosity
   // settings precedence over report_id.
   function automatic void set_verbosity(ovm_component comp, string report_id);
      ovm_object             tmp;
      verbosity_t            tmp_verb;

      string  comp_name;

      if (comp == null)
        `ovm_fatal(report_id, "Component passed into set_verbosity is null.")
      else
        comp_name = comp.get_full_name();

      // --- Set verbosity level by report_id ---
      // Set the verbosity level in the calling component from the value
      // in the report_id lookup table (if found)
      if (verbosity.size() != 0) begin // Save the associative array lookup
                                       // if the array is empty
         if (verbosity.exists(report_id)) begin
            tmp_verb = verbosity[report_id];
            if (tmp_verb.recur) begin
               `ovm_info(report_id,
                  $sformatf("Recursively setting verbosity to %0d for %s",
                            tmp_verb.level,
                            report_id),
                  OVM_DEBUG)
               if (comp.get_num_children() == 0) begin
                  // If component has no children, set verbosity
                  // non-hierarchically. Uncomment the following
                  // warning for debugging hierarchy problems
                  comp.set_report_verbosity_level(tmp_verb.level);
                  //`ovm_warning(report_id, $sformatf("component %s has no children. Recursion setting will have no effect.  Ensure you are calling set_verbosity after building all the children.", comp_name))
               end // if (comp.get_num_children() == 0)
               else begin
                  comp.set_report_verbosity_level_hier(tmp_verb.level);
               end // else begin
            end // if (tmp_verb.recur)
            else begin
               `ovm_info(report_id,
                  $sformatf("Setting verbosity to %0d for %s",
                            tmp_verb.level,
                            report_id),
                  OVM_DEBUG)
               comp.set_report_verbosity_level(tmp_verb.level);
            end // else begin
         end // if (verbosity.exists(report_id))
      end // if (verbosity.size() != 0)

      // --- Set verbosity level by hierarchical name ---
      // Set the verbosity level in the calling component from the value
      // in the hierarchical-name lookup table (if found)
      if (hverbosity.size() != 0) begin // Save the associative array lookup
                                        // if the array is empty
         // Get hierarchical name of the calling component
         string  comp_name = comp.get_full_name();

         if (hverbosity.exists(comp_name)) begin
            tmp_verb = hverbosity[comp_name];
            if (tmp_verb.recur) begin
               `ovm_info(report_id,
                  $sformatf({"Recursively setting verbosity to ",
                             "%0d for hierarchical name %s"},
                            tmp_verb.level,
                            report_id),
                  OVM_DEBUG)
               if (comp.get_num_children() == 0) begin
                  // If component has no children, set verbosity
                  // non-hierarchically. Uncomment the following
                  // warning for debugging hierarchy problems
                  comp.set_report_verbosity_level(tmp_verb.level);
                  //`ovm_warning(report_id, $sformatf("component %s has no children. Recursion setting will have no effect.  Ensure you are calling set_verbosity after building all the children.", comp_name))
               end // if (comp.get_num_children() == 0)
               else begin
                  comp.set_report_verbosity_level_hier(tmp_verb.level);
               end // else begin
            end // if (tmp_verb.recur) begin
            else begin
               `ovm_info(report_id,
                  $sformatf("Setting verbosity to %0d for hierarchical name %s",
                            tmp_verb.level,
                            report_id),
                  OVM_DEBUG)
               comp.set_report_verbosity_level(tmp_verb.level);
            end // else begin
         end // if (hverbosity.exists(comp_name)) begin
      end // if (hverbosity.size() != 0)
   endfunction : set_verbosity

endpackage : runtime_verbosity_pkg

