//
//  Copyright 2010 Xilinx, Inc.
//
//  Licensed under the Apache License, Version 2.0 (the "License");
//  you may not use this file except in compliance with the License.
//  You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
//  Unless required by applicable law or agreed to in writing, software
//  distributed under the License is distributed on an "AS IS" BASIS,
//  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//  See the License for the specific language governing permissions and
//  limitations under the License.
//
//  -----------------------------------------------------------------------
//
//  File:        runtime_verbosity_server_pkg.sv
//  Revision:    1.0, 2010-Nov-05
//  Description: Package to provide runtime, component-level control over
//               verbosity in an OVM testbench. This package also requires
//               the accompanying runtime_verbosity_pkg.
//  Usage:
//    Import this package in your TB (preferrably the top-level). See the
//    command-line options described in runtime_verbosity_pkg.sv. Note
//    that when using this package, a call to
//    runtime_verbosity_pkg::set_verbosity() is not _required_ in each
//    component's build() function, but if set_verbosity() is not called AND
//    the `ovm_info macro is used, the verbosity for that component can only
//    be decreased, not increased.
//
`include "ovm_macros.svh"

package runtime_verbosity_server_pkg;

   import ovm_pkg::*;
   import runtime_verbosity_pkg::verbosity;

   // verbosity_ovm_report_server is a custom ovm_report_server which
   // overrides the "report" call, allowing the verbosity to be overridden
   // based on the report_id, but still uses the OVM code for displaying
   // reports
   class verbosity_ovm_report_server extends ovm_report_server;

      static verbosity_ovm_report_server global_override_server = null;

      function new();
         super.new();
         set_global_server();
      endfunction : new

      // Report function - modifies the calling component's verbosity setting
      // based on the lookup-table, then calls the built-in report function,
      // and finally restores the verbosity setting before returning
      function void report(
         ovm_severity severity,
         string name,
         string id,
         string message,
         int verbosity_level,
         string filename,
         int line,
         ovm_report_object client
         );

         ovm_report_handler rh;
         int saved_verbosity_level;

         // Get the report handler from the calling component and save off its
         // original verbosity level
         rh = client.get_report_handler();
         saved_verbosity_level = rh.m_max_verbosity_level;

         // Check the lookup table to see if there is an entry that matches
         // the passed-in id. If so, override the calling component's verbosity
         // setting with the value from the lookup-table
         if (verbosity.exists(id)) begin
           rh.m_max_verbosity_level = verbosity[id].level;
         end

         // Call the OVM built-in report function, which will now "see" the
         // verbosity setting from the lookup-table
         super.report(severity, name, id, message, verbosity_level,
                      filename, line, client);

         // Restore the original verbosity setting before returning
         rh.m_max_verbosity_level = saved_verbosity_level;

      endfunction : report

      // Function to override the global report server with our custom one
      // which knows about the verbosity lookup-table.
      virtual function void set_global_server();
         ovm_report_global_server global_server;
         ovm_object tmp;

         if (global_override_server == null) begin
            global_override_server = this;

            global_server = new();          // Get the global server
            global_server.set_server(this); // Override the global server
         end
      endfunction : set_global_server

   endclass : verbosity_ovm_report_server

   // Create singleton utility class which override the default report server
   verbosity_ovm_report_server verb_server = new();

endpackage : runtime_verbosity_server_pkg

