////////////////////////////////////////////////////////////////////////////
// mem_driver_c - the unified sequence driver
////////////////////////////////////////////////////////////////////////////
class mem_driver_c extends ovm_driver #(mem_item_c, mem_item_c);

   const string report_id = "MEM_DRIVER";
   
   // ports
   //Using the REQ type specified from the ovm_driver parameterization
   ovm_analysis_port #(REQ)  sb_out_ap;
   
   // stimulus handle
   REQ                       transaction_h;

   //Done bit
   bit                       done = 1;

   //Config object handle
   mem_config_c              mem_cfg_h;
   
   // virtual interface
   virtual mem_interface     mem_interface_h;
   
   `ovm_component_utils(mem_driver_c)
   
   function new(string name, ovm_component parent);
      super.new(name, parent);
      //Enables using the stop task to delay simulation end
      enable_stop_interrupt = 1;
   endfunction : new
   
   function void build();
      super.build();
      
      //Use the runtime verbosity package
      runtime_verbosity_pkg::set_verbosity(.comp(this), .report_id(report_id));
      
      //Get the config object
      mem_cfg_h = mem_config_c::get_config(this, "mem_config");
      
      // new the analysis port
      sb_out_ap = new("sb_out_ap", this);
   endfunction : build
   
   function void connect();
      mem_interface_h = mem_cfg_h.mem_interface;
   endfunction : connect
   
   task run();
      `ovm_info(report_id,  "running unified_seq driver",OVM_HIGH )
      // set everything to zero before the reset
      mem_interface_h.address           <= '0;
      mem_interface_h.data_to_dut       <= '0;
      mem_interface_h.data_to_dut_valid <=  0;         
      mem_interface_h.latency           <= '0;
      mem_interface_h.instruction       <=  PAUSE;
      
      // on reset .....
      @(negedge mem_interface_h.reset);
      forever @(posedge mem_interface_h.clock) begin
         
         // Tell the unified_seq that we need a new transaction
         seq_item_port.get_next_item(transaction_h);
         done = 0;
         
         // wait for a while if it is a pause 
         if (transaction_h.instruction == PAUSE) begin
            done = 1;
            mem_interface_h.instruction       =  PAUSE;
            for (int i = 0; i < transaction_h.latency;i++) begin
               `ovm_info(report_id, "pausing",OVM_HIGH )
               @(posedge mem_interface_h.clock);                     
            end
            // otherwise pass the instruction to the DUT through the intf
         end else begin    
            // using non-blocking assignments here because the assignments
            // are to interface signals which are static, even though this 
            // is dynamic code it is necessary to use non-blocking 
            // assignments here to ensure that the values do not change 
            // before the monitor has sampled them!
            // assign the pins through the interface from the transaction
            mem_interface_h.address     <= transaction_h.address;
            mem_interface_h.data_to_dut <= transaction_h.data_to_dut; 
            mem_interface_h.latency     <= transaction_h.latency;
            mem_interface_h.instruction <= transaction_h.instruction;
            // set the pin to tell the dut that the data is ready
            mem_interface_h.data_to_dut_valid <= 1;         
            // reset all the values ready for the next transaction
            @(negedge mem_interface_h.dut_busy);
            // send out any transactions
            transaction_h.data_from_dut = mem_interface_h.data_from_dut;
            sb_out_ap.write(transaction_h);                    
            `ovm_info(report_id,
                            $sformatf("sending an instruction out of the analysis port Instruction: %s, Address: %d Data: %d", transaction_h.instruction, transaction_h.address, transaction_h.data_from_dut), OVM_HIGH)
            done = 1;
         end // else: !if(transaction_h.instruction == PAUSE)
         seq_item_port.item_done();
         
         // reset the bus
         mem_interface_h.data_to_dut_valid <=  0;
         mem_interface_h.address           <= '0;
         mem_interface_h.data_to_dut       <= '0;
         mem_interface_h.latency           <= '0;         
         mem_interface_h.instruction       <=  PAUSE;
         
      end     // forever       
   endtask : run

   task stop(string ph_name);
      fork
         wait (done == 1);
         #10000;
      join_any
   disable fork;
   endtask : stop
   
endclass : mem_driver_c
