//////////////////////////////////////////////////////////////////////////////
// stimulus_c - the basic transaction(stimulus) object.
//////////////////////////////////////////////////////////////////////////////
class mem_item_c extends ovm_sequence_item;

   const string report_id = "MEM_ITEM";

                rand instruction_type        instruction;
                rand logic            [ 7:0] address;
                rand logic            [15:0] data_to_dut;
                rand logic                   data_to_dut_valid;
                rand logic            [ 7:0] latency;
                     logic            [15:0] data_from_dut;
          local rand int                     choose_read_address;
   static local      logic            [15:0] addresses_written_list [int];
   static local      int                     instructions_sent;

   //Register with the Factory
   `ovm_object_utils(mem_item_c)
     
/*   //This is how you would do it without the macros 
   typedef ovm_component_registry #(mem_item_c, "mem_item_c") type_id;

   static function type_id get_type();
      return type_id::get();
   endfunction // string

   function string get_type_name();
      return "mem_item_c";
   endfunction*/

   constraint latency_cycle        { latency == 1                               ;}
   // this is functionally unimportant as the DUT doesn't read it but it 
   // makes the waveforms easier to read
   constraint instruction_first    { solve instruction before data_to_dut       ;}
   constraint write_data_on_read   { if (instruction == READ)  data_to_dut == '0 ;}
   constraint dont_write_to_zero   { if (instruction == WRITE) address     != '0 ;}

   // note: this is safe becuase address 0 returns the number of accesses so
   // is valid from the start
   constraint valid_read_address { choose_read_address >= 0 ; 
                                   choose_read_address <= instructions_sent ; }

   function new(string name = "mem_item_c");
      super.new(name);
      // the 0 address returns the number of accesses so is always valid
      // this is initialized here in case the first instruction is a read.
      addresses_written_list[0] = 0;
   endfunction : new

   
   // Need to ensure that reads are random but are only from addresses that
   // have been written to! 
   function void post_randomize();
      string message;
      // record the write addresses in an associative array
      if (instruction == WRITE) begin
         instructions_sent++;
         $sformat(message,"WRITE: instruction number %0d, address 0x%02x", instructions_sent, address);
         `ovm_info(report_id, message, OVM_DEBUG)
         addresses_written_list[instructions_sent] = address;
      end
      // for the read address, chose it from the array (addresses_written_list) written during write cycles
      if (instruction == READ) begin
         $sformat(message,"READ selected %0d address written, address 0x%02x", choose_read_address, addresses_written_list[choose_read_address]);
         `ovm_info(report_id, message, OVM_DEBUG )
         address = addresses_written_list[choose_read_address];
      end
   endfunction : post_randomize

   
   //In order to get debug or status information printed to a simulator
   // transcript or a log file from a data object there needs to be a way to 
   // convert the objects contents into a string representation - this is the 
   // purpose of the convert2string() method. Calling the method will return a
   // string detailing the values of each of the properties formatted for 
   // displaying or writing to a file. The format is determined by the user.
   function string convert2string();
      // Note you could use \t (tab) and \n (newline) to format the data in
      // columns.  The enumerated op_code types .name() method returns a
      // string corresponding to its value
      return $sformatf("Address: 0x%02x, Data(to): 0x%04x,  Latency: %0d, Instruction: %s", address, data_to_dut, latency, instruction.name());
   endfunction : convert2string


   //The do_print() method is called by the ovm_object print() method. Its 
   // original purpose was to print out a string representation of an ovm data 
   // object using one of the ovm_printer policy classes. However, a higher 
   // performance version of the same functionality can be achieved by 
   // implementing the method as a wrapper around a $display() call that takes 
   // the string returned by a convert2string() call as an argument.
   function void do_print(ovm_printer printer);
      if(printer.knobs.sprint == 0) begin
         $display(convert2string());
      end
      else begin
         printer.m_string = convert2string();
      end 
   endfunction: do_print

   
   //The do_compare method is called by the ovm_object compare() method and it
   // is used to compare two data objects of the same type to determine 
   // whether their contents are equal. The do_compare() method should only be 
   // coded for those properties which can be compared.
   //The ovm_comparer policy object has to be passed to the do_compare() 
   // method for compatability with the virtual method template, but it is not 
   // necessary to use it in the comparison function and performance can be 
   // improved by not using it.
   function bit do_compare(ovm_object rhs, ovm_comparer comparer);
      mem_item_c rhs_;
      
      //If the cast fails, comparison has also failed
      //A check for null is not needed because that is done in the compare()
      // function which calls do_compare()
      if (!$cast(rhs_, rhs)) return 0;
      
      //This is a bit contrived because it compares a write cycle to a read 
      // cycle so the data is on a different bus.  It would probably be better
      // to move to just "data" at this level and let the driver work out if 
      // it is to or from at the pin level
      return (super.do_compare(rhs, comparer) &&
              (address == rhs_.address) &&
              (data_from_dut == rhs_.data_to_dut));
   endfunction : do_compare
             

   //The purpose of the do_copy method is to provide a means of making a deep 
   // copy of a data object.  The do_copy method is either used on its own or 
   // via the ovm_objects clone() method which allows independent duplicates 
   // of a data object to be created.
   //Note that the rhs argument is of type ovm_object since it is a virtual 
   // method, and that it therefore needs to be cast to the actual transaction
   // type before its fields can be copied. A design consideration for this 
   // method is that it may not always make sense to copy all property values 
   // from one object to another.
   function void do_copy(ovm_object rhs);
      mem_item_c rhs_;

      if(!$cast(rhs_, rhs)) begin
         `ovm_error(report_id, "Cast failed in do_copy()");
         return;
      end
      
      super.do_copy(rhs); // Chain the copy with parent classe
      instruction = rhs_.instruction;
      address = rhs_.address;
      data_to_dut = rhs_.data_to_dut;
      data_to_dut_valid = rhs_.data_to_dut_valid;
      latency = rhs_.latency;
      data_from_dut = rhs_.data_from_dut;
   endfunction :do_copy

   /* Use the default clone
   function ovm_object clone();
      mem_item_c t = new this;  //Shallow Copy
      `ovm_info(report_id, "cloning", OVM_LOW)
      return t;
   endfunction : clone*/


   //The do_report() method is intended to support the viewing of data objects 
   // as transactions in a waveform GUI. Like the printing data object 
   // methods, the principle is that the fields that are recorded are visible 
   // in the transaction viewer. The underlying implementation of the 
   // do_record() method is simulator specific and for Questa involves the use 
   // of the $add_attribute() system call.
   //In order to get transaction viewing to work with Questa you need to:
   //  * Implement the do_report() method as shown
   //  * Use the Questa ovm closed kit - i.e. The compiled version of OVM 
   //    shipped with Questa
   //  * Set the recording_detail config item to OVM_FULL in your test or env
   //    set_config_int("*", "recording_detail", OVM_FULL);
   
   // Macro to help with recording - Questa specific
   `define ovm_record_field(NAME,VALUE) \
      $add_attribute(recorder.tr_handle, VALUE, NAME);
   
   // do_record:
   function void do_record(ovm_recorder recorder);
      super.do_record(recorder); // To record any inherited data members
      `ovm_record_field("instruction", instruction.name())
      `ovm_record_field("address", address)
      `ovm_record_field("data_to_dut", data_to_dut)
      `ovm_record_field("latency", latency)
      `ovm_record_field("data_from_dut", data_from_dut)
   endfunction : do_record
   
endclass : mem_item_c


//Classes used in Unified Sequence 3 / Scenario 3
class mem_item_latency4_change_c extends mem_item_c;
   `ovm_object_utils(mem_item_latency4_change_c)
   constraint latency_cycle        { latency == 4 ;}
endclass : mem_item_latency4_change_c

class mem_item_latency2_change_c extends mem_item_c;
   `ovm_object_utils(mem_item_latency2_change_c)
   constraint latency_cycle        { latency == 2 ;}
endclass : mem_item_latency2_change_c

class mem_item_latency6_change_c extends mem_item_c;
   `ovm_object_utils(mem_item_latency6_change_c)
   constraint latency_cycle        { latency == 6 ;}
endclass : mem_item_latency6_change_c

//Classes used in Unifed Sequence 4 / Scenario 4
class mem_item_latency_random_c extends mem_item_c;
   `ovm_object_utils(mem_item_latency_random_c)
   constraint latency_cycle        { latency <  7;
                                     latency != 0; }
endclass : mem_item_latency_random_c
